__all__ = ['TIMED_INPUT_MATCHER', 'to_bdd2']

import re

import attr
import aiger
import aiger_bv as BV
import aiger_bdd
import funcy as fn
from aiger import common as cmn

try:
    from dd.cudd import BDD
except ImportError:
    from dd.autoref import BDD

from mce.order import BitOrder


TIMED_INPUT_MATCHER = re.compile(r'(.*)##time_(\d+)\[(\d+)\]')
INPUT_MATCHER = re.compile(r'(.*)\[(\d+)\]')


def cone(circ: BV.AIGBV, output: str) -> BV.AIGBV:
    """Return cone of influence aigbv output."""
    for out in circ.outputs - {output}:
        size = circ.omap[out].size
        circ >>= BV.sink(size, [out])
    assert len(circ.outputs) == 1
    assert fn.first(circ.outputs) == output
    return circ


def to_bdd2(mdp, horizon, output=None, manager=None):
    if output is None:
        assert len(mdp.outputs) == 1
        output = fn.first(mdp.outputs)

    # 1. Convert MDP to AIGBV circuit.
    circ = cone(mdp.aigbv, output)

    # 2. Define Causal Order
    inputs, env_inputs = mdp.inputs, circ.inputs - mdp.inputs
    imap = circ.imap

    def count_bits(inputs):
        return sum(imap[i].size for i in inputs)

    order = BitOrder(count_bits(inputs), count_bits(env_inputs), horizon)

    # 4. Create topological ordering on variables.
    step, old2new_lmap = circ.cutlatches()
    init = dict(old2new_lmap.values())
    init = step.imap.blast(init)
    states = set(init.keys())

    def flattened(t):
        def fmt(k):
            idxs = range(imap[k].size)            
            return [f"{k}##time_{t}[{i}]" for i in idxs]

        actions = fn.lmapcat(fmt, inputs)
        coin_flips = fn.lmapcat(fmt, env_inputs)
        return actions + coin_flips

    unrolled_inputs = fn.lmapcat(flattened, range(horizon))

    levels = {k: i for i, k in enumerate(unrolled_inputs)}

    # 3. Create BDD

    manager = BDD() if manager is None else manager
    manager.declare(*levels.keys())
    manager.reorder(levels)
    manager.configure(reordering=False)

    def bdd_bool(x):
        return manager.true if x else manager.false

    gate_nodes = {aiger.aig.Input(k): bdd_bool(v) for k, v in init.items()}
    state_inputs = [aiger.aig.Input(k) for k in init.keys()]

    for time in range(horizon):
        # Only remember states.
        gate_nodes = fn.project(gate_nodes, state_inputs)

        for gate in cmn.eval_order(step.aig):
            if isinstance(gate, aiger.aig.ConstFalse):
                gate_nodes[gate] = manager.false
            elif isinstance(gate, aiger.aig.Inverter):
                gate_nodes[gate] = ~gate_nodes[gate.input]
            elif isinstance(gate, aiger.aig.Input):
                if gate.name in states:
                    continue

                name, idx = INPUT_MATCHER.match(gate.name).groups()
                name = f'{name}##time_{time}[{idx}]'
                gate_nodes[gate] = manager.var(name)

            elif isinstance(gate, aiger.aig.AndGate):
                gate_nodes[gate] = gate_nodes[gate.left] & gate_nodes[gate.right]

        for si in state_inputs:
            gate_nodes[si] = gate_nodes[step.aig.node_map[si.name]]

    assert step.omap[output].size == 1
    output = step.omap[output][0]

    return gate_nodes[step.aig.node_map[output]], manager, order



__all__ = ['spec2graph']

from typing import Optional

import attr
import funcy as fn
import networkx as nx
from bdd2dfa.b2d import to_dfa, BNode

from mce.spec import ConcreteSpec


def spec2graph(spec: ConcreteSpec, qdd=False, dprob=None) -> nx.DiGraph:
    dfa = spec._as_dfa(qdd=qdd)

    if dprob is None:
        dprob = lambda *_: None

    def is_sink(state) -> bool:
        return state.node.var is None

    def is_decision(state) -> bool:
        lvl = state.node.level
        if qdd:
            lvl -= state.debt
        return spec.order.is_decision(lvl)

    def key(state):
        return (state.ref, state.debt) if qdd else state.ref

    sinks = set()
    g = nx.DiGraph()

    @fn.memoize
    def _node(state):
        decision = is_decision(state) and not is_sink(state)
        lvl = state.node.level
        var = state.label() if state.node.var is None else state.node.var
        node = key(state)
        g.add_node(node, lvl=lvl, var=var, decision=decision)
        return node

    stack = [dfa.start]
    while len(stack) > 0:
        state = stack.pop()
        action2succ = {a: dfa._transition(state, a) for a in dfa.inputs}

        for action, succ in action2succ.items():
            if key(succ) not in g.nodes:
                stack.append(succ)

            if succ == state:  # Sink
                sinks.add(_node(succ))
                continue                

            if qdd and state.debt > 0:
                g.add_edge(_node(state), _node(succ), action=None, prob=1)
            else:
                if is_decision(state):
                    prob = dprob(key(state), action)
                else:
                    prob = 1/2

                g.add_edge(_node(state), _node(succ), action=action, prob=prob)

    g.add_node("DUMMY", lvl=None, var=None, decision=False)
    for sink in sinks:
        g.add_edge(sink, "DUMMY", action=None, prob=1)

    g = nx.freeze(g)    
    return g, _node(dfa.start), list(sinks)



__all__ = ["preimage"]

import operator as op
from functools import reduce
from typing import TypeVar, Mapping, Tuple, List, Iterable



# import aiger_coins as C
import aiger_bv as BV


CircVal = Mapping[str, Tuple[bool]]
AtomicPred = CircVal
AtomicPreds = List[AtomicPred]
BVExpr = BV.UnsignedBVExpr


def gen_equiv_checks(aps: AtomicPreds) -> Iterable[BVExpr]:
    for t, ap in enumerate(aps):
        for name, val in ap.items():
            size = len(val)
            timed_name = f"{name}##time_{t + 1}"
            sym = BV.atom(size, timed_name, signed=False)
            sym_val = BV.atom(size, val, signed=False)

            yield sym == sym_val


def preimage(aps: AtomicPreds, mdp: BV.AIGBV, is_unrolled=False) -> BVExpr:
    """
    Returns a circuit which checks if an action sequence results in a
    given sequence of observations/atomic predicates (aps).
    """
    assert len(aps) > 1

    unrolled = mdp if is_unrolled else mdp.unroll(len(aps)) 
        
    check_val = reduce(op.and_, gen_equiv_checks(aps)).aigbv

    test = unrolled >> check_val
    assert test.inputs == unrolled.inputs
    assert len(test.outputs) == 1
    
    return BVExpr(test)



from typing import Sequence

import funcy as fn
import numpy as np
from bdd2dfa.b2d import BNode
from scipy.special import logsumexp

from mce.policy3 import BitPolicy


def coin_flips(n) -> Sequence[bool]:
    return 


def sample(ctrl):
    node = ctrl.root
    while node != "DUMMY":
        kids = ctrl.graph.neighbors(node)
        probs = [ctrl.prob(node, k) for k in kids]
        node2 = np.random.choice(kids, p=probs)

        delta = graph.nodes[node2]['lvl'] - graph.nodes[node]['lvl']
        if ctrl.graph.nodes[node]['decision']:
            delta -= 1  # Use one bit
            yield ctrl.graph.edges[(node, node2)]['action']

        yield from coin_flips(delta)
        node = node2



import time
from multiprocess import Pool

import funcy as fn

from mce.policy3 import fit
from mce.spec import concretize
from mce.demos import encode_trcs, log_likelihoods


def spec_mle(mdp, demos, specs, top=100, parallel=False, psat=None):
    """
    Searches for the most likely specification in specs given
    demonstations, demos, from an agent operating in mdp.
    """
    horizon = len(demos[0][0])
    specs = list(specs)

    start_time = time.time()
    print("encoding traces")
    encoded_trcs = encode_trcs(mdp, demos)
    print(f"done encoding traces")

    @fn.memoize
    def score(spec):
        start_time = time.time()
        times = {}

        print("concretizing spec")
        cspec = concretize(spec, mdp, horizon)
        print("done spec")
        times["build spec"] = time.time() - start_time

        if psat is None:
            sat_prob = sum(cspec.accepts(trc) for trc in encoded_trcs)
            sat_prob /= len(encoded_trcs)
        else:
            sat_prob = psat

        start_time = time.time()
        print("fitting policy")
        ctrl = fit(cspec, sat_prob)
        print(f"done fitting")
        times["fit"] = time.time() - start_time

        start_time = time.time()
        print(f"compute log likelihood of demos")
        lprob = log_likelihoods(ctrl, encoded_trcs)

        times["surprise"] = time.time() - start_time

        print("\n----------------------------\n")
        print(f"BDD size: {cspec.bexpr.dag_size}")
        print(f"Controller Size: {ctrl.size}")
        print(f"log_prob: {lprob}")
        print("\n".join(f"{key}: {val:.2}s" for key, val in times.items()))
        print("\n----------------------------\n")

        print(times)

        return lprob

    if parallel:
        _specs = list(enumerate(specs))

        def score2(spec):
            i, spec = spec
            return i, score(spec)

        spec2score = dict(Pool().map(score2, _specs))
        spec2score = fn.walk_keys(lambda idx: specs[idx], spec2score)
        best_spec = max(specs, key=spec2score.get)

    else:
        best_spec = max(specs, key=score)
        spec2score = fn.walk_keys(lambda x: x[0], score.memory)
    return best_spec, spec2score



__all__ = ['visitation_graph', 'encode_trcs', 'log_likelihoods']

from collections import Counter
from typing import Tuple, List, Sequence, Mapping

import funcy as fn
import networkx as nx
import numpy as np
from bdd2dfa.b2d import BNode
from scipy.special import logsumexp

from mce.policy3 import BitPolicy
from mce.spec import ConcreteSpec


Demo = List[Mapping[str, bool]]
Demos = List[Demo]

AbstractDemo = Sequence[Mapping[str, bool]]
AbstractDemos = Sequence[AbstractDemo]

Edge = Tuple[int, int]


def encode_trcs(dyn, trcs):
    """Encodes i/o traces as sequences of (sys, env) actions."""
    return [_encode_trc(dyn, *v) for v in trcs]


def _encode_trc(dyn, sys_actions, states):
    return dyn.encode_trc(sys_actions, states)


def log_likelihoods(ctrl: BitPolicy, demos: Demos) -> float:
    dfa = ctrl.spec._as_dfa(qdd=True)

    def trace(actions):
        bits = ctrl.spec.flatten(actions)
        trc = ((s.ref, s.debt) for s in dfa.trace(bits))
        return zip(trc, bits)

    # Convert to sequence of bdd, action pairs.
    io_seq = fn.chain(fn.mapcat(trace, demos))
    return sum(ctrl.prob(n, a, log=True, qdd=True) for n, a in io_seq)


def node2action_hist(spec: ConcreteSpec, demos: Demos):
    dfa = spec._as_dfa(qdd=True)

    def trace(actions):
        bits = spec.flatten(actions)
        trc = ((s.ref, s.debt) for s in dfa.trace(bits))
        return zip(trc, bits)

    traces = fn.mapcat(trace, demos)
    node2actions = fn.group_values(traces)
    return fn.walk_values(Counter, node2actions)


def visitation_graph(ctrl: BitPolicy, demos: Demos):
    observed_bias = node2action_hist(ctrl.spec, demos)

    graph, root, real_sinks = ctrl.markov_chain()
    for node in nx.topological_sort(graph):
        for node2 in graph.neighbors(node):
            action = graph.edges[node, node2]['action']

            if (node not in observed_bias):
                visitation = 0
            else:
                if action is not None:
                    visitation = observed_bias[node][action]
                else:
                    visitation = observed_bias[node][False]
                    visitation += observed_bias[node][True]

            graph.edges[node, node2]['visitation'] = visitation / len(demos)

    for sink in real_sinks:
        incoming = graph.in_edges(sink)
        visitation = sum(graph.edges[e]['visitation'] for e in incoming)
        graph.edges[sink, "DUMMY"]['visitation'] = visitation

    return graph, root, real_sinks



from itertools import product

import aiger
import aiger_bv as BV
import aiger_coins as C
import aiger_ptltl as PLTL
import funcy as fn
import hypothesis.strategies as st


def sys1():
    return C.circ2mdp(BV.identity_gate(1, 'a'))


def sys2():
    mdp = sys1()
    mdp <<= C.coin((1, 8), name='c')
    mdp >>= C.circ2mdp(BV.sink(1, ['c']))  # HACK
    return mdp


def sys3():
    mdp = sys1()
    mdp |= C.circ2mdp(BV.tee(1, {'c': ['c', 'c_next']}))
    coin = (~C.coin((1, 2), name='c')).with_output('c')
    mdp <<= coin

    delay = BV.aig2aigbv(aiger.delay(['c'], [True]))
    delay = C.circ2mdp(delay)

    return mdp >> delay


SPEC1 = PLTL.atom('a').historically()
SPEC2 = PLTL.atom(True)
SPEC3 = (PLTL.atom('a') == PLTL.atom('c')).historically()


def gen_scenario(spec, sys):
    def scenario(negate=False):
        spec2 = ~spec if negate else spec
        return spec2, sys

    return scenario


scenario1 = gen_scenario(
    spec=SPEC1, sys=sys1(),
)


scenario2 = gen_scenario(
    spec=SPEC2, sys=sys1(),
)

scenario_reactive = gen_scenario(
    spec=SPEC3, sys=sys3(),
)


def make_strategy(spec_sys):
    spec, sys = spec_sys
    return st.just(gen_scenario(spec=spec, sys=sys()))


DET_SCENARIOS = st.one_of(
    fn.lmap(make_strategy, product([SPEC1, SPEC2], [sys1, sys2]))
)

NOT_DET_SCENARIOS = st.one_of(
    fn.lmap(make_strategy, product([SPEC3, SPEC2], [sys3]))
)

SCENARIOS = st.one_of(DET_SCENARIOS, NOT_DET_SCENARIOS)



__all__ = ['BitOrder']

from typing import Tuple

import attr


@attr.s(frozen=True, auto_attribs=True)
class BitOrder:
    decision_bits: int
    chance_bits: int
    horizon: int

    @property
    def total_bits(self) -> int:
        return self.decision_bits + self.chance_bits

    def round_index(self, lvl: int) -> int:
        return lvl % self.total_bits

    def is_decision(self, lvl_ctx) -> bool:
        lvl = lvl_ctx if isinstance(lvl_ctx, int) else lvl_ctx.curr_lvl
        return self.round_index(lvl) < self.decision_bits

    def time_interval(self, lvl: int) -> Tuple[int]:
        idx = self.round_index(lvl)
        return lvl - idx, lvl + (self.total_bits - 1 - idx)

    def interval(self, lvl: int) -> Tuple[int]:
        bot, top = self.time_interval(lvl)
        boundary = bot + self.decision_bits - 1

        if self.is_decision(lvl):
            return bot, boundary
        return boundary + 1, top

    def time_step(self, lvl):
        return lvl // self.total_bits

    def skipped_time_steps(self, lvl1, lvl2):
        if lvl2 < lvl1:
            return -self.skipped_time_steps(lvl2, lvl1)

        t1_prev, t2_prev = map(self.time_step, (lvl1, lvl2))
        return max(0, t2_prev - t1_prev - 1)

    def skipped_decisions(self, lvl1, lvl2) -> int:
        if lvl2 < lvl1:
            return -self.skipped_decisions(lvl2, lvl1)
        elif lvl2 - lvl1 <= 1:
            return 0
        elif self.total_bits == 1:
            assert self.decision_bits == 1
            return lvl2 - lvl1 - 1

        itvl1, itvl2 = map(self.time_interval, (lvl1 + 1, lvl2 - 1))
        if itvl1 == itvl2:
            return self._rrd(lvl1 + 1, lvl2 - 1)

        skipped_rounds = self.skipped_time_steps(lvl1, lvl2)

        start = (lvl1 >= itvl1[0])*self._rrd(lvl1 + 1, itvl1[1])
        middle = self.decision_bits * skipped_rounds
        end = (lvl2 <= itvl2[1])*self._used_decisions(lvl2 - 1)
        return start + middle + end

    def _rrd(self, lvl1, lvl2):
        """Remaining Round Decisions"""
        if not self.is_decision(lvl1):
            return 0

        idx1, idx2 = map(self.round_index, (lvl1, lvl2))
        idx2 = min(idx2, self.decision_bits - 1)
        return idx2 - idx1 + 1

    def _used_decisions(self, lvl):
        """Remaining Round Decisions"""
        return min(self.decision_bits, self.round_index(lvl) + 1)



__all__ = ["BitPolicy", "policy", "fit"]

import random
from functools import partial
from typing import Tuple, List, Optional, Hashable, Mapping

import attr
import funcy as fn
import networkx as nx
import numpy as np
import scipy as sp
from bidict import bidict
from scipy.special import logsumexp
from scipy.optimize import brentq

from mce.spec import ConcreteSpec
from mce.nx import spec2graph


@attr.s(frozen=True, auto_attribs=True)
class BitPolicy:
    spec: ConcreteSpec
    ref2action_dist: Mapping[int, Mapping[bool, float]]
    lsat: float

    @property
    def psat(self) -> float:
        """
        Probability of satisfying the underlying concrete
        specification using this policy.
        """
        return np.exp(self.lsat)

    @property
    def size(self) -> int:
        """
        Number of BDD nodes with non-uniform defined action
        distributions.
        """
        return len(self.ref2action_dist)
        
    def prob(self, node, action, log=False, qdd=False) -> float:
        """Probability of agent applying action given bdd node."""
        if action is None:
            prob = 1
        else:
            if qdd:
                node, debt = node
            else:
                debt = 0

            if debt > 0:
                prob =  0.5
            else:
                uniform = {True: 0.5, False: 0.5}
                prob = self.ref2action_dist.get(node, uniform)[action]
        return np.log(prob) if log else prob

    def markov_chain(self):
        def dprob(ref_debt, action):
            ref, debt = ref_debt
            return 1 if debt > 0 else self.prob(ref, action)

        return spec2graph(self.spec, qdd=True, dprob=dprob)

    def _simulate(self, seed: int = None, copolicy=False):
        np.random.seed(seed)
        graph, node, _ = self.markov_chain()

        while graph.out_degree(node) > 0:
            kids = list(graph.neighbors(node))
            assert len(kids) == graph.out_degree(node)

            probs = np.array([graph.edges[node, k]['prob'] for k in kids])

            if copolicy:
                probs = 1 - probs
            
            kid, *_ = random.choices(kids, weights=probs)

            # Pr(s' | s, a)
            action = graph.edges[node, kid]['action']

            node = kid
            if node == "DUMMY":
                continue
            elif action is None:
                action = random.choice([True, False])

            yield action

    def simulate(self, seed: int = None, copolicy=False):
        """
        Generates tuples of (state, action, next_state) and the
        probability transitioning from state to next_state.
        """
        bits = self._simulate(seed, copolicy)
        chunks = fn.chunks(self.spec.order.total_bits, bits)
        return [self.spec.unflatten(c)[0] for c in chunks]

    def stochastic_matrix(self):
        """
        Returns underlying markov chain as stochastic_matrix and a map
        nodes to index.

        If the matrix more than one node, e.g., the concrete spec is
        not constant, then:

        - The false node is index 0.
        - The true node is index 1.
        - The root (start of episode) node is index 2.

        Note that true and false have self loops added to make the
        matrix stochastic.
        """
        graph, root, real_sinks = self.markov_chain()
        assert len(real_sinks) > 0

        if len(real_sinks) == 1:
            mat = sp.sparse.csr_matrix((1, 1))
            mat[0, 0] = 1
            return mat, bidict(enumerate(real_sinks))

        var = lambda x: graph.nodes(data=True)[x]['var']
        false, true = sorted(real_sinks, key=var)

        # Get nodes in correct order.
        nodes = ["DUMMY", false, true, root]
        nodes += list(set(graph.nodes) - set(nodes))

        mat = nx.adjacency_matrix(graph, nodes, weight="prob")
        mat = mat[1:, 1:]

        # Make sink node self loop in stochastic matrix.
        mat[0, 0] = 1
        mat[1, 1] = 1

        return mat, bidict(enumerate(nodes[1:]))


def policy(spec: ConcreteSpec, coeff: Optional[float] = None) -> BitPolicy:
    reference_graph, root, real_sinks = spec2graph(spec)

    @fn.cache(5)  # Cache results for 5 seconds.
    def ppolicy(coeff):
        graph = reference_graph.reverse(copy=True)

        node_data = graph.nodes(data=True)

        var = lambda x: graph.nodes(data=True)[x]['var']
        lvl = lambda x: graph.nodes(data=True)[x]['lvl']
        decision = lambda x: graph.nodes(data=True)[x]['decision']

        # Iterate in reverse topological order (ignoring dummy sink).
        for node in fn.rest(nx.topological_sort(graph)):
            if isinstance(var(node), bool):
                graph.nodes[node]['val'] = coeff*int(var(node))                
                graph.nodes[node]['lsat'] = 0 if var(node) else -float('inf')
                continue
                
            kids = [c for (c, _) in graph.in_edges(node)]
            vals = np.array([graph.nodes[c]['val'] for c in kids])
            lsats = np.array([graph.nodes[c]['lsat'] for c in kids])

            if not decision(node):
                probs = np.array([graph[c][node]['prob'] for c in kids])
                graph.nodes[node]['val'] = (probs * vals).sum()
            else:
                # Account for skipped decisions.
                # Note: This is arguably a bug in the model.
                skipped = np.array([
                    spec.order.skipped_decisions(lvl(node), lvl(c)) for c in kids
                ])
                vals = vals + skipped * np.log(2)

                state_val = graph.nodes[node]['val'] = logsumexp(vals)
                probs = np.exp(vals - state_val)

                # Add action_probs to edges. 
                for prob, child in zip(probs, kids):
                    graph[child][node]['prob'] = prob

            graph.nodes[node]['lsat'] = logsumexp(lsats, b=probs)
        
        def decision_nodes():
            return (n for n in graph.nodes if graph.nodes[n]['decision'])

        def action_dist(node):
            assert graph.nodes[node]['decision']
            out_edges = (graph.edges[e] for e in graph.in_edges(node))

            return {
                data['action']: data['prob'] for data in out_edges
            }

        ref2adist = {n: action_dist(n)  for n in decision_nodes()}
        
        return BitPolicy(
            spec=spec, 
            lsat=graph.nodes[root]['lsat'],
            ref2action_dist=ref2adist
        )

    if coeff is None:
        return ppolicy

    return ppolicy(coeff)


def fit(cspec: ConcreteSpec, psat: float, top: float=100) -> BitPolicy:
    """Fit a max causal ent policy with satisfaction probability psat."""
    assert 0 <= psat <= 1
    pctrl = policy(cspec)

    def f(coeff):
        return pctrl(coeff).psat - psat

    if f(-top) > 0:
        coeff = 0
    elif f(top) < 0:
        coeff = top
    else:
        coeff = brentq(f, -top, top)

    if coeff < 0:
        # More likely the negated spec than this one.
        coeff = 0  

    return pctrl(coeff)



import networkx as nx


def label_edge(graph, edge):
    data = graph.edges()[edge]

    # Note: Sometimes entries purposely set to None.
    label = ""
    if data.get('prob') is not None:
        label += f'p={data["prob"]:.2f}\n'
    if data.get('action') is not None:
        label += f'a={data["action"]}\n'
    if data.get('visitation') is not None:
        label += f'f={data["visitation"]}\n'

    data['label'] = label


def label_node(graph, node):
    data = graph.nodes()[node]
    if data['decision'] or isinstance(data['var'], bool):
        data['label'] = f"var={data['var']}\nlvl={data['lvl']}"


def draw(graph, path):
    for edge in graph.edges():
        label_edge(graph, edge)

    for node in graph.nodes():
        label_node(graph, node)

    # convert to a graphviz agraph
    nx.drawing.nx_pydot.write_dot(graph, path)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mce']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=19.3,<20.0',
 'bdd2dfa>=0.4.0,<0.5.0',
 'bidict>=0.18.3,<0.19.0',
 'dfa[draw]>=2.0.0,<3.0.0',
 'funcy>=1.13,<2.0',
 'multiprocess>=0.70.9,<0.71.0',
 'py-aiger-bdd>=0.3.1,<0.4.0',
 'py-aiger-bv>=2.0.0,<3.0.0',
 'py-aiger-coins[sat,bdd]>=1.4.0,<2.0.0',
 'py-aiger-sat>=1.1.0,<2.0.0',
 'scipy>=1.4.1,<2.0.0']

setup_kwargs = {
    'name': 'mce-spec-inference',
    'version': '0.1.0',
    'description': 'Maximum Causal Entropy Specification Inference.',
    'long_description': '',
    'author': 'Marcell Vazquez-Chanlatte',
    'author_email': 'mvc@linux.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mvcisback/mce-spec-inference',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



__all__ = ["ConcreteSpec", "concretize"]

from typing import FrozenSet, Sequence, Mapping, Tuple

import attr
import aiger_bdd
import aiger_bv as BV
import aiger_coins as C
import funcy as fn
from aiger_bv.bundle import BundleMap
from bdd2dfa.b2d import to_dfa, BNode
from bidict import bidict
from dfa import DFA

from mce.preimage import preimage
from mce.order import BitOrder
from mce.bdd import to_bdd2, TIMED_INPUT_MATCHER


Action = Mapping[str, Sequence[bool]]
Actions = Sequence[Action]
Bits = Sequence[bool]


def xor(x, y):
    return (x | y) & ~(x & y)


@attr.s(frozen=True, auto_attribs=True)
class ConcreteSpec:
    """
    Models an concrete specification over sequences of
    system/environment action pairs encoded as bitvectors.
    """

    bexpr: "BDD"
    order: BitOrder  # TODO: Make this a derived quantity.
    dyn: BV.AIGBV
    sys_inputs: FrozenSet[str] = attr.ib(converter=frozenset)

    @property
    def imap(self) -> BundleMap:
        """System input map."""
        return self.dyn.imap.project(self.sys_inputs)

    @property
    def emap(self) -> BundleMap:
        """Environment input map."""
        return self.dyn.imap.omit(self.sys_inputs)

    @property
    def horizon(self) -> int:
        return self.order.horizon

    @property
    def manager(self):
        return self.bexpr.bdd

    def flatten(self, actions: Actions) -> Bits:
        """
        Converts structured sequence of (sys, env) actions to a
        sequence of bits that this concrete specification recognizes.
        """
        manager = self.manager

        timed_actions = {}
        bmap = self.imap + self.emap
        for t, action in enumerate(actions):
            old2new = {k: f'{k}##time_{t}' for k in bmap.keys()}
            bmap_t = bmap.relabel(old2new)
            action_t = fn.walk_keys(old2new.get, action)
            timed_actions.update(bmap_t.blast(action_t))

        idx2key = bidict(self.bexpr.bdd.vars).inv
        return [timed_actions[idx2key[i]] for i in range(len(idx2key))]

    def unflatten(self, bits: Bits) -> Actions:
        """
        Take a sequence of bits and group into bitvector inputs for
        dynamics circuit.
        """
        assert (len(bits) % self.order.total_bits) == 0
        return list(self._unflatten(bits))

    def _unflatten(self, bits: Bits):
        size = self.order.total_bits
        for i, chunk in enumerate(fn.chunks(self.order.total_bits, bits)):
            mapping = {}
            for j, bit in enumerate(chunk):
                lvl = i*size + j
                var = self.bexpr.bdd.var_at_level(lvl)
                name, _, idx = TIMED_INPUT_MATCHER.match(var).groups()
                mapping[f'{name}[{idx}]'] = bit

            yield self.dyn.imap.unblast(mapping)
        
    def accepts(self, actions: Actions) -> bool:
        """Does this spec accept the given sequence of (sys, env) actions."""
        flattened = self.flatten(actions)
        assert len(flattened) == self.order.horizon * self.order.total_bits
        return self._as_dfa(qdd=True).label(flattened)

    def toggle(self, actions: Actions):
        """Toggles a sequence of (sys, env) actions."""
        assert len(actions) == self.horizon
        aps = fn.lpluck(0, self.dyn.simulate(actions))
        expr = preimage(aps=aps, mdp=self._unrolled(), is_unrolled=True)
        bexpr, *_ = aiger_bdd.to_bdd(
            expr, manager=self.manager, renamer=lambda _, x: x
        )

        return attr.evolve(self, bexpr=xor(self.bexpr, bexpr))

    @fn.cache(60 * 5)  # Evict after 5 minutes.
    def _unrolled(self) -> BV.AIGBV:
        return self.dyn.unroll(self.horizon)

    @fn.cache(60 * 1)  # Evict after 1 minute.
    def _as_dfa(self, qdd=False) -> DFA:
        """
        Returns a dfa with binary alphabet which models the
        ConcreteSpecification with the order given by self.bexpr.
        """
        return to_dfa(self.bexpr, qdd=qdd)

    def abstract_trace(self, actions: Actions) -> Sequence[BNode]:
        """Path a set of (sys, env) actions takes through BDD."""
        return self._as_dfa(qdd=True).trace(self.flatten(actions))


def concretize(
        monitor, sys: C.MDP, horizon: int, manager=None
) -> ConcreteSpec:
    """
    Convert an abstract specification monitor and a i/o transition
    system into a concrete specification over the horizion.
    """
    # Make format correct.
    if not isinstance(monitor, C.MDP):
        assert hasattr(monitor, 'aig') or hasattr(monitor, 'aigbv')
        if hasattr(monitor, 'aigbv'):
            monitor = monitor.aigbv
        else:
            monitor = BV.aig2aigbv(monitor.aig)
        monitor = C.MDP(monitor)

    # Remove ignored outputs of sys.
    for sym in (monitor.inputs ^ sys.outputs):
        size = sys._aigbv.omap[sym].size
        monitor >>= C.MDP(BV.sink(size, [sym]))    

    bexpr, manager, order = to_bdd2(sys >> monitor, horizon)
    return ConcreteSpec(bexpr, order, sys_inputs=sys.inputs, dyn=sys.aigbv)
